@@ -58,6 +58,8 @@
 #include <sys/stat.h>
 #include <sys/wait.h>
 
+#include <gio/gunixinputstream.h>
+#include <gio/gunixoutputstream.h>
 #include <glib.h>
 #include <glib/gi18n.h>
 #include <glib-object.h>
@@ -125,6 +127,8 @@ static gchar* control_directory = NULL;
 static guint timeout_id = 0;
 static gboolean initialization_completed = FALSE;
 static GMainLoop *loop = NULL;
+static int parent_wakeup_fd = -1;
+static GDBusConnection *system_bus_connection = NULL;
 
 static GOptionEntry option_entries[] = {
 	{ "start", 's', 0, G_OPTION_ARG_NONE, &run_for_start,
@@ -605,22 +609,43 @@ discover_other_daemon (DiscoverFunc callback, gboolean acquire)
 }
 
 static void
+redirect_fds_after_fork (void)
+{
+	int fd, i;
+
+	for (i = 0; i < 3; ++i) {
+		fd = open ("/dev/null", O_RDONLY);
+		sane_dup2 (fd, i);
+		close (fd);
+	}
+}
+
+static void
+block_on_fd (int fd)
+{
+	unsigned char dummy;
+	read (fd, &dummy, 1);
+}
+
+static int
 fork_and_print_environment (void)
 {
 	int status;
 	pid_t pid;
-	int fd, i;
+	int wakeup_fds[2] = { -1, -1 };
 
 	if (run_foreground) {
-		print_environment ();
-		return;
+		return -1;
 	}
 
+	if (!g_unix_open_pipe (wakeup_fds, FD_CLOEXEC, NULL))
+		exit (1);
+
 	pid = fork ();
 
 	if (pid != 0) {
-
 		/* Here we are in the initial process */
+		close (wakeup_fds[1]);
 
 		if (run_daemonized) {
 
@@ -633,8 +658,8 @@ fork_and_print_environment (void)
 				exit (WEXITSTATUS (status));
 
 		} else {
-			/* Not double forking */
-			print_environment ();
+			/* Not double forking, wait for child */
+			block_on_fd (wakeup_fds[0]);
 		}
 
 		/* The initial process exits successfully */
@@ -654,6 +679,7 @@ fork_and_print_environment (void)
 		pid = fork ();
 
 		if (pid != 0) {
+			close (wakeup_fds[1]);
 
 			/* Here we are in the intermediate child process */
 
@@ -665,7 +691,7 @@ fork_and_print_environment (void)
 				exit (1);
 
 			/* We've done two forks. */
-			print_environment ();
+			block_on_fd (wakeup_fds[0]);
 
 			/* The intermediate child exits */
 			exit (0);
@@ -674,12 +700,7 @@ fork_and_print_environment (void)
 	}
 
 	/* Here we are in the resulting daemon or background process. */
-
-	for (i = 0; i < 3; ++i) {
-		fd = open ("/dev/null", O_RDONLY);
-		sane_dup2 (fd, i);
-		close (fd);
-	}
+	return wakeup_fds[1];
 }
 
 static gboolean
@@ -807,6 +828,163 @@ on_login_timeout (gpointer data)
 	return FALSE;
 }
 
+static void
+on_vanished_quit_loop (GDBusConnection *connection,
+                       const gchar *name,
+                       gpointer user_data)
+{
+	g_main_loop_quit (user_data);
+}
+
+static void on_logind_session_property_get (GObject *connection,
+					    GAsyncResult *res,
+					    gpointer user_data G_GNUC_UNUSED)
+{
+	GError *error = NULL;
+	GVariant *result, *resultv;
+	const gchar *state;
+	gboolean should_quit;
+
+	result = g_dbus_connection_call_finish (G_DBUS_CONNECTION (connection), res, &error);
+
+	if (error) {
+		if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+			g_critical ("%s Couldn't get session state: %s", G_STRLOC, error->message);
+		g_error_free (error);
+		return;
+	}
+
+	g_variant_get (result, "(v)", &resultv, NULL);
+	state = g_variant_get_string (resultv, NULL);
+
+	should_quit = g_strcmp0 (state, "closing") == 0;
+
+	g_clear_pointer (&result, g_variant_unref);
+	g_clear_pointer (&resultv, g_variant_unref);
+
+	/* yes, the session is closing, so we'll quit now */
+	if (should_quit)
+		cleanup_and_exit (0);
+}
+
+static void on_logind_session_properties_changed (GDBusConnection *connection,
+						  const gchar *sender_name G_GNUC_UNUSED,
+						  const gchar *object_path,
+						  const gchar *interface_name G_GNUC_UNUSED,
+						  const gchar *signal_name G_GNUC_UNUSED,
+						  GVariant *parameters,
+						  gpointer user_data G_GNUC_UNUSED)
+{
+	const gchar *prop_iface;
+	gboolean active;
+	GVariant* changed_properties;
+
+	g_variant_get (parameters, "(&s@a{sv}^as)", &prop_iface, &changed_properties, NULL);
+
+	if (g_variant_lookup (changed_properties, "Active", "b", &active, NULL)) {
+		if (!active) {
+			/* ok, the session went inactive, let's see if that is because
+			 * it is closing */
+			g_dbus_connection_call (
+				connection,
+				"org.freedesktop.login1",
+				object_path,
+				"org.freedesktop.DBus.Properties",
+				"Get",
+				g_variant_new ("(ss)", prop_iface, "State"),
+				G_VARIANT_TYPE ("(v)"),
+				G_DBUS_CALL_FLAGS_NONE,
+				-1,
+				NULL,
+				on_logind_session_property_get,
+				NULL
+			);
+		}
+	}
+
+	g_variant_unref (changed_properties);
+}
+
+static void
+on_logind_object_path_get (GObject *connection,
+			   GAsyncResult *res,
+			   gpointer user_data G_GNUC_UNUSED)
+{
+	GError *error = NULL;
+	GVariant *result;
+	const gchar *object_path;
+	gchar *remote_error;
+	gboolean is_cancelled, is_name_has_no_owner;
+
+	result = g_dbus_connection_call_finish (G_DBUS_CONNECTION (connection), res, &error);
+
+	/* If there's an error we always want to quit - but we only tell the
+	 * user about it if something went wrong. Cancelling the operation or
+	 * not having logind available are okay. */
+	if (error) {
+		is_cancelled = g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED);
+
+		remote_error = g_dbus_error_get_remote_error (error);
+		is_name_has_no_owner = g_strcmp0 (remote_error, "org.freedesktop.DBus.Error.NameHasNoOwner") == 0;
+
+		if (!is_cancelled && !is_name_has_no_owner)
+			g_critical ("%s Couldn't get object path: %s", G_STRLOC, error->message);
+
+		g_free (remote_error);
+		g_error_free (error);
+		return;
+	}
+
+	/* now we know which object path to look on, watch for
+	 * PropertiesChanged. Note that, per logind's documentation, we only
+	 * get notified for 'Active' changing */
+	g_variant_get (result, "(&o)", &object_path, NULL);
+
+	g_dbus_connection_signal_subscribe (
+		G_DBUS_CONNECTION (connection),
+		"org.freedesktop.login1",
+		"org.freedesktop.DBus.Properties",
+		"PropertiesChanged",
+		object_path,
+		NULL,
+		G_DBUS_SIGNAL_FLAGS_NONE,
+		on_logind_session_properties_changed,
+		NULL,
+		NULL
+	);
+
+	g_clear_pointer (&result, g_variant_unref);
+}
+
+static void
+start_watching_logind_for_session_closure ()
+{
+	g_return_if_fail (system_bus_connection != NULL);
+
+	const gchar *xdg_session_id;
+
+	xdg_session_id = g_getenv ("XDG_SESSION_ID");
+
+	if (!xdg_session_id)
+		return;
+
+	/* get the right object path */
+	g_dbus_connection_call (
+		system_bus_connection,
+		"org.freedesktop.login1",
+		"/org/freedesktop/login1",
+		"org.freedesktop.login1.Manager",
+		"GetSession",
+		g_variant_new ("(s)", xdg_session_id, NULL),
+		G_VARIANT_TYPE ("(o)"),
+		G_DBUS_CALL_FLAGS_NO_AUTO_START,
+		-1,
+		NULL,
+		on_logind_object_path_get,
+		NULL
+	);
+}
+
 int
 main (int argc, char *argv[])
 {
@@ -830,6 +1008,9 @@ main (int argc, char *argv[])
 	 * predictable startup.
 	 */
 
+	GDBusConnection *connection = NULL;
+	GError *error = NULL;
+
 	/*
 	 * Before we do ANYTHING, we drop privileges so we don't become
 	 * a security issue ourselves.
@@ -876,16 +1057,37 @@ main (int argc, char *argv[])
 		exit (0);
 	}
 
+	if (perform_unlock) {
+		login_password = read_login_password (STDIN);
+		atexit (clear_login_password);
+	}
+
+	/* The whole forking and daemonizing dance starts here. */
+	parent_wakeup_fd = fork_and_print_environment();
+
 	/* The --start option */
 	if (run_for_start) {
 		if (discover_other_daemon (initialize_daemon_at, TRUE)) {
 			/*
-			 * Another daemon was initialized, print out environment
-			 * for any callers, and quit or go comatose.
+			 * Another daemon was initialized, print out environment,
+			 * tell parent we're done, and quit or go comatose.
 			 */
 			print_environment ();
-			if (run_foreground)
-				while (sleep(0x08000000) == 0);
+			close (parent_wakeup_fd);
+			if (run_foreground) {
+				connection = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
+				if (error) {
+					g_warning ("Couldn't connect to session bus: %s", error->message);
+					g_clear_error (&error);
+				}
+				loop = g_main_loop_new (NULL, FALSE);
+				g_bus_watch_name (G_BUS_TYPE_SESSION, "org.gnome.keyring",
+				                  G_BUS_NAME_WATCHER_FLAGS_NONE,
+				                  NULL, on_vanished_quit_loop, loop, NULL);
+				g_main_loop_run (loop);
+				g_clear_pointer (&loop, g_main_loop_unref);
+				g_clear_object (&connection);
+			}
 			cleanup_and_exit (0);
 		}
 
@@ -908,11 +1110,6 @@ main (int argc, char *argv[])
 	if (!gkd_control_listen ())
 		return FALSE;
 
-	if (perform_unlock) {
-		login_password = read_login_password (STDIN);
-		atexit (clear_login_password);
-	}
-
 	/* The --login option. Delayed initialization */
 	if (run_for_login) {
 		timeout_id = g_timeout_add_seconds (LOGIN_TIMEOUT, (GSourceFunc) on_login_timeout, NULL);
@@ -924,10 +1121,23 @@ main (int argc, char *argv[])
 			cleanup_and_exit (1);
 	}
 
+	/* if we can get a connection to the system bus, watch it and then kill
+	 * ourselves when our session closes */
+
+	system_bus_connection = g_bus_get_sync (G_BUS_TYPE_SYSTEM, NULL, NULL);
+
+	if (system_bus_connection)
+		start_watching_logind_for_session_closure ();
+
 	signal (SIGPIPE, SIG_IGN);
 
-	/* The whole forking and daemonizing dance starts here. */
-	fork_and_print_environment();
+	/* Print the environment and tell the parent we're done */
+	print_environment ();
+
+	if (!run_foreground) {
+		close (parent_wakeup_fd);
+		redirect_fds_after_fork ();
+	}
 
 	g_unix_signal_add (SIGTERM, on_signal_term, loop);
 	g_unix_signal_add (SIGHUP, on_signal_term, loop);
